import { QueriesObserver, notifyManager } from '@tanstack/query-core';
import { flushSync, onDestroy, onMount, unstate, untrack } from 'svelte';
import { useIsRestoring } from './useIsRestoring';
import { useQueryClient } from './useQueryClient';
import { createMemo, createResource } from './utils.svelte';
export function createQueries({ queries, ...options }, queryClient) {
    const client = useQueryClient(queryClient);
    const isRestoring = useIsRestoring();
    const queriesStore = $derived(typeof queries != 'function' ? () => queries : queries);
    const defaultedQueriesStore = createMemo(() => {
        return queriesStore().map((opts) => {
            const defaultedOptions = client.defaultQueryOptions(opts);
            // Make sure the results are already in fetching state before subscribing or updating options
            defaultedOptions._optimisticResults = isRestoring
                ? 'isRestoring'
                : 'optimistic';
            return defaultedOptions;
        });
    });
    const observer = new QueriesObserver(client, defaultedQueriesStore(), options);
    $effect(() => {
        // Do not notify on updates because of changes in the options because
        // these changes should already be reflected in the optimistic result.
        observer.setQueries(defaultedQueriesStore(), options, { listeners: false });
    });
    let result = $state(observer.getOptimisticResult(defaultedQueriesStore())[1]());
    /*   $effect.pre(() => {
      const unsubscribe = isRestoring
        ? () => undefined
        : observer.subscribe(
            notifyManager.batchCalls((v) => {
              result = v
            }),
          )
  
      return () => unsubscribe()
    }) */
    //
    $effect(() => {
        if (queries.length) {
            untrack(() => {
                result = observer.getOptimisticResult(defaultedQueriesStore())[1]();
            });
        }
    });
    const dataResources = $derived(result.map((queryRes) => {
        const dataPromise = () => new Promise((resolve) => {
            if (queryRes.isFetching && queryRes.isLoading)
                return;
            resolve(unstate(queryRes.data));
        });
        return createResource(dataPromise);
    }));
    flushSync(() => {
        for (let index = 0; index < dataResources.length; index++) {
            const dataResource = dataResources[index];
            dataResource[1].mutate(() => unstate(result[index].data));
            dataResource[1].refetch();
        }
    });
    let taskQueue = [];
    const subscribeToObserver = () => observer.subscribe((result_) => {
        taskQueue.push(() => {
            flushSync(() => {
                const dataResources_ = dataResources;
                for (let index = 0; index < dataResources_.length; index++) {
                    const dataResource = dataResources_[index];
                    const unwrappedResult = { ...unstate(result_[index]) };
                    result[index] = unstate(unwrappedResult);
                    dataResource[1].mutate(() => unstate(result_[index].data));
                    dataResource[1].refetch();
                }
            });
        });
        queueMicrotask(() => {
            const taskToRun = taskQueue.pop();
            if (taskToRun)
                taskToRun();
            taskQueue = [];
        });
    });
    let unsubscribe = () => undefined;
    $effect.pre(() => {
        unsubscribe = isRestoring ? () => undefined : subscribeToObserver();
        // cleanup needs to be scheduled after synchronous effects take place
        return () => queueMicrotask(unsubscribe);
    });
    onDestroy(unsubscribe);
    onMount(() => {
        observer.setQueries(defaultedQueriesStore(), undefined, {
            listeners: false,
        });
    });
    const handler = (index) => ({
        get(target, prop) {
            if (prop === 'data') {
                return dataResources[index][0]();
            }
            return Reflect.get(target, prop);
        },
    });
    const getProxies = $derived(() => result.map((s, index) => {
        return new Proxy(s, handler(index));
    }));
    const proxifiedState = $state(getProxies());
    /*  $effect(() => {
      console.log(
        'result updated',
        result,
        JSON.stringify(result),
        JSON.stringify(proxifiedState),
      )
    })
    $effect(() => {
      console.log(
        'proxifiedState',
  
        JSON.stringify(proxifiedState),
      )
    })*/
    $effect.pre(() => {
        if (getProxies) {
            untrack(() => {
                Object.assign(proxifiedState, getProxies());
            });
        }
    });
    return proxifiedState;
}
7;
